/**
   Page 369 of the textbook. (Fixed a bug in line 6 of their code.)

   The recurrence relation is
      r(n) = max( p[i] + r[n-i] : with 1 <= i <= min(n, p.length) )
*/
class CutRod_Memoized
{
   public static int cutRod(int[] p, int n)
   {
      final int[] r = new int[n+1]; // We will remember the solution values in r.
      for (int i = 0; i <= n; ++i)
         r[i] = -1; // Flag that denotes i'th problem hasn't been solved yet.

      return cutRod(p, n, r);
   }


   public static int cutRod(int[] p, int n, int[] r)
   {
      if (r[n] >= 0) // Already have a solution for length n?
      {
         return r[n];
      }

      int q;
      if (n == 0)
      {
         q = 0;
      }
      else
      {
         q = Integer.MIN_VALUE;
         for (int i = 1; i <= Math.min(n, p.length-1); ++i) // i is the position of the first cut
         {
            q = Math.max( q, p[i] + cutRod(p, n-i, r) );  // recursion
         }
      }
      r[n] = q; // Remember the solution value for length n.
      return q;
   }


   public static void main(String[] args)
   {
      //                   1  2  3  4   5   6   7   8   9  10
      final int[] p = {-1, 1, 5, 8, 9, 10, 17, 17, 20, 24, 30};

      for (int i = 1; i <= 35; ++i)
      {
         System.out.println( "CutRod_memoized("+i+") = " + cutRod(p, i) );
      }
   }
}