
import java.util.List;
import java.util.ArrayList;

/**
   Compute the actual solution, not the value of the solution.

   The recurrence relation is
      r(n) = max( p[i] + r[n-i] : with 1 <= i <= min(n, p.length) )
*/
class CutRod_Memoized_v2
{
   public static List<Integer> cutRod(int[] p, int n)
   {
      final List<List<Integer>> r = new ArrayList<>(); // We will remember the solution values in r.
      for (int i = 0; i <= n; ++i)
         r.add(null);  // Flag that denotes i'th problem hasn't been solved yet.

      return cutRod(p, n, r);
   }


   public static List<Integer> cutRod(int[] p, int n, List<List<Integer>> r)
   {
      if (r.get(n) != null) // Already have a solution for length n?
      {
         return r.get(n);
      }

      List<Integer> soln;
      if (n == 0)
      {
         soln = new ArrayList<>();
      }
      else
      {
         soln = new ArrayList<>();
         for (int i = 1; i <= Math.min(n, p.length-1); ++i) // i is the position of the first cut
         {
            final List<Integer> tempSoln = new ArrayList<>( cutRod(p, n-i, r) );  // recursion
            tempSoln.add(0, i);
            if ( value(tempSoln, p) > value(soln, p) )
            {
               soln = tempSoln;
            }
         }
      }
      r.set(n, soln); // Remember the solution value for length n.
      return soln;
   }

   public static int value(List<Integer> soln, int[] p)
   {
      int value = 0;
      for (Integer i : soln)
      {
         value += p[i];
      }
      return value;
   }


   public static void main(String[] args)
   {
      //                   1  2  3  4   5   6   7   8   9  10
      final int[] p = {-1, 1, 5, 8, 9, 10, 17, 17, 20, 24, 30};

      for (int i = 1; i <= 35; ++i)
      {
         final List<Integer> soln = cutRod(p, i);
         System.out.println( "CutRod("+i+") = " + soln + " with value = " + value(soln, p));
      }
   }
}