
import java.util.List;
import java.util.ArrayList;
import java.util.Arrays;

/**
   Compute the structure of a solution to
   the longest increasing subsequence
   dynamic programming problem.

      lis(a) = max { LE(a, j) }
             0<=j<a.length
   where

      LE(a, j) = 1 + max { LE(a, i) }
                    i < j
                 a[i] < a[j]
*/
public class LIS_v2
{
   /**
      Compute the structure of a solution to the
      longest increasing subsequence using the
      helper function LE(a, j).

      Encode the structure of a solution as a List
      of pairs (a[n], n) where a[n] is an entry from
      the array that is in the longest increasing
      subsequence and n is that entry's index in
      the array.

      @param a  array of integers
      @return a List that encodes the entries in the longest increasing subsequence
   */
   public static List<List<Integer>> lis(int[] a)
   {
      List<List<Integer>> soln = new ArrayList<>();

      for (int j = 0; j < a.length; ++j)
      {
         final List<List<Integer>> soln2 = LE1(a, j);

         if (soln2.size() > soln.size())
            soln = soln2;
      }

      return soln;
   }


   /**
      Compute the structure of a longest increasing
      subsequence from a that ends at entry a[j].

      This is a purely recursive (and very inefficient,
      exponential) implementation of LE(a, j).

      @param a  array of integers
      @param j  index of where the increasing subsequence ends
      @return a List that encodes the entries in the longest increasing subsequence that ends at index j
   */
   public static List<List<Integer>> LE1(int[] a, int j)
   {
      List<List<Integer>> soln = new ArrayList<>();
      if ( j == 0 )
      {
        soln.add( Arrays.asList(new Integer[]{a[0], 0}) );
      }
      else
      {
         for (int i = 0; i < j; ++i)
         {
            if ( a[i] < a[j] )
            {
               final List<List<Integer>> soln2 = LE1(a,i);
               if (soln2.size() > soln.size())
                  soln = soln2;
            }
         }
         soln.add( Arrays.asList(new Integer[]{a[j], j}) );
      }
      return soln;
   }



   public static void main(String[] args)
   {
      final int START = 5;
      final int END   = 20;

      for (int n = START; n <= END; ++n)
      {
         int[] a = new int[n];

         // Initialize the array with the numbers for 1 to n.
         for (int j = 0; j < a.length; ++j) a[j] = j+1;

         // The permutation() method will randomly
         // permute the contents of the array.
         LIS.permutation(a, 0, a.length - 1);

         // If n is small enough, print out the array.
         if (a.length <= 50)
         {
            System.out.println(java.util.Arrays.toString(a));
         }

         // Compute the solution of the
         // longest increasing subsequence
         // in the suffled array.
         final List<List<Integer>> soln = lis(a);
         System.out.printf("lis(%d) = " + soln + ".\n", a.length);
      }
   }
}
